# Generated by Django 4.2.11 on 2024-03-26 10:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0008_alter_activity_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="is_public",
            field=models.BooleanField(
                default=True, help_text="If True, all internal users can access this account", verbose_name="Public"
            ),
        ),
    ]
